//
//  MWAudioDecoder.m
//  videoGL
//
//  Created by Romain Criton on 28/07/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "MWAudioDecoder.h"
#import <MobileCoreServices/MobileCoreServices.h>

#include "libavformat/avformat.h"
#include <mach/mach_time.h>

#pragma mark Audio Queue Services code
#define kNumberBuffers 3
typedef struct AQPlayerState
{
    AudioStreamBasicDescription   mDataFormat;
    AudioQueueRef                 mQueue;
    AudioQueueBufferRef           mBuffers[kNumberBuffers];
    MWAudioDecoder                *decoder;
    UInt32                        bufferByteSize;
    SInt64                        mCurrentPacket;
    UInt32                        mNumPacketsToRead;
    AudioStreamPacketDescription  *mPacketDescs;
    BOOL                          mIsRunning;
} AQPlayerState;

void DeriveBufferSize (
					   AudioStreamBasicDescription *ASBDesc,                            // 1
					   UInt32                      maxPacketSize,                       // 2
					   Float64                     seconds,                             // 3
					   UInt32                      *outBufferSize,                      // 4
					   UInt32                      *outNumPacketsToRead                 // 5
					   )
{
    static const int maxBufferSize = 0x50000;                        // 6
    static const int minBufferSize = 0x4000;                         // 7
	
    if (ASBDesc->mFramesPerPacket != 0) {                             // 8
        Float64 numPacketsForTime =
		ASBDesc->mSampleRate / ASBDesc->mFramesPerPacket * seconds;
        *outBufferSize = numPacketsForTime * maxPacketSize;
    } else {                                                         // 9
        *outBufferSize =
		maxBufferSize > maxPacketSize ?
		maxBufferSize : maxPacketSize;
    }
	
    if (                                                             // 10
        *outBufferSize > maxBufferSize &&
        *outBufferSize > maxPacketSize
		)
        *outBufferSize = maxBufferSize;
    else {                                                           // 11
        if (*outBufferSize < minBufferSize)
            *outBufferSize = minBufferSize;
    }
	
    *outNumPacketsToRead = *outBufferSize / maxPacketSize;           // 12
}

static void HandleOutputBuffer (
								void                *aqData,
								AudioQueueRef       inAQ,
								AudioQueueBufferRef inBuffer
								) {
    AQPlayerState *pAqData = (AQPlayerState *) aqData;
    if (!pAqData->mIsRunning) return;
	
	MWAudioDecoder *decoder = pAqData->decoder;
    UInt32 numBytesReadFromFile = 0;
    UInt32 numPackets = pAqData->mNumPacketsToRead;
	for(int i = 0; i < numPackets; i++)
	{
		MWMediaPacket *packet = [decoder.inputBuffer dequeueSignalIfLow];
		if(packet == nil)
		{
			numPackets = i;
			break;
		}
		
		AVPacket *avp = (AVPacket *)packet.data;
		void *rawPacket = avp->data;
		int byteSize = avp->size;
		memcpy(inBuffer->mAudioData + numBytesReadFromFile, rawPacket, byteSize);
		
		if(pAqData->mPacketDescs != NULL)
		{
			AudioStreamPacketDescription *desc = &((pAqData->mPacketDescs)[i]);
			desc->mStartOffset = numBytesReadFromFile;
			desc->mVariableFramesInPacket = 0;
			desc->mDataByteSize = byteSize;
		}
		numBytesReadFromFile += byteSize;
	}
    if (numPackets > 0) {
        inBuffer->mAudioDataByteSize = numBytesReadFromFile;
		AudioQueueEnqueueBuffer ( 
								 pAqData->mQueue,
								 inBuffer,
								 (pAqData->mPacketDescs ? numPackets : 0),
								 pAqData->mPacketDescs
								 );
        pAqData->mCurrentPacket += numPackets;
		DLog(@"MWAudioDecoder: adding %d packets", numPackets);
    } else {
        AudioQueueStop (
						pAqData->mQueue,
						false
						);
        pAqData->mIsRunning = NO; 
    }
}

void audioTimeStampDifference(AudioTimeStamp *diff, AudioTimeStamp *start, AudioTimeStamp *end)
{
	if((diff == NULL) || (start == NULL) || (end == NULL)) return;
	
	diff->mFlags = 0;
	
	if((start->mFlags & end->mFlags & kAudioTimeStampSampleTimeValid) != 0)
	{
		diff->mSampleTime = end->mSampleTime - start->mSampleTime;
		diff->mFlags |= kAudioTimeStampSampleTimeValid;
	}
	
	if((start->mFlags & end->mFlags & kAudioTimeStampHostTimeValid) != 0)
	{
		diff->mHostTime = end->mHostTime - start->mHostTime;
		diff->mFlags |= kAudioTimeStampHostTimeValid;
	}
}

@interface MWAudioDecoder()
-(void)setStartAudioTimeStamp;
-(void)getCurrentAudioTimeStamp:(AudioTimeStamp *)ts;

@end


@implementation MWAudioDecoder

@synthesize inputBuffer = _inputBuffer;

-(id)initWithPath:(NSString *)path inputBuffer:(FIFOStack *)buf
{
	if (!(self=[super init])) return nil;
	
	// Alloc Audio Queue state machine
	_state = malloc(sizeof(AQPlayerState));
	AQPlayerState *pState = (AQPlayerState *)_state;

	// Get an Audio Queue file handle
	AudioFileID file;
	CFURLRef audioFileURL = CFURLCreateFromFileSystemRepresentation (
											 NULL,
											 (const UInt8 *) [path cStringUsingEncoding:NSASCIIStringEncoding],                       // 3
											 strlen ([path cStringUsingEncoding:NSASCIIStringEncoding]),                              // 4
											 false
											 );
	OSStatus result = AudioFileOpenURL (
					  audioFileURL,
					  kAudioFileReadPermission,
					  0,
					  &file
					  );
	CFRelease (audioFileURL);
	if(result < 0) goto initError;
	
	// Fill data format information
	UInt32 dataFormatSize = sizeof (pState->mDataFormat);
	result = AudioFileGetProperty (file,
								   kAudioFilePropertyDataFormat,
								   &dataFormatSize,
								   &(pState->mDataFormat));
	if(result < 0)
	{
		pState->mDataFormat.mSampleRate		=	44100.0;
		pState->mDataFormat.mFormatID			=	kAudioFormatMPEG4AAC;
		pState->mDataFormat.mFormatFlags		=	0;
		pState->mDataFormat.mBytesPerPacket	=	0;
		pState->mDataFormat.mFramesPerPacket	=	1024;
		pState->mDataFormat.mBytesPerFrame		=	0;
		pState->mDataFormat.mChannelsPerFrame	=	2;
		pState->mDataFormat.mBitsPerChannel	=	0;
		pState->mDataFormat.mReserved			=	0;
	}
	
	// Fill max packet size
	UInt32 maxPacketSize;
	UInt32 propertySize = sizeof (maxPacketSize);
	result = AudioFileGetProperty (file,
								   kAudioFilePropertyPacketSizeUpperBound,
								   &propertySize,
								   &maxPacketSize
								   );
	if(result < 0)
	{
		maxPacketSize = 500;
	}
	
	// Compute required buffer size
	DeriveBufferSize (&(pState->mDataFormat),
					  maxPacketSize,
					  0.5,
					  &pState->bufferByteSize,
					  &pState->mNumPacketsToRead);
	
	// Create Audio Queue
	result = AudioQueueNewOutput (&pState->mDataFormat,
								  HandleOutputBuffer,
								  pState,
								  CFRunLoopGetCurrent (),
								  kCFRunLoopCommonModes,
								  0,
								  &pState->mQueue);
	if(result < 0) goto initError;
	
	// Additionnal config in case of VBR format
	bool isFormatVBR = (pState->mDataFormat.mBytesPerPacket == 0 ||
						pState->mDataFormat.mFramesPerPacket == 0);
	if (isFormatVBR) {
		pState->mPacketDescs =
		(AudioStreamPacketDescription*) malloc (pState->mNumPacketsToRead * sizeof (AudioStreamPacketDescription));
	} else {
		pState->mPacketDescs = NULL;
	}
	
	// Get Magic Cookie
	UInt32 cookieSize = sizeof (UInt32);
	bool couldNotGetProperty = AudioFileGetPropertyInfo (
								file,
								kAudioFilePropertyMagicCookieData,
								&cookieSize,
								NULL
							  );
	if (!couldNotGetProperty && cookieSize)
	{
		char* magicCookie =
        (char *) malloc (cookieSize);
		
		result = AudioFileGetProperty (file,
							  kAudioFilePropertyMagicCookieData,
							  &cookieSize,
							  magicCookie);
		result = AudioQueueSetProperty (pState->mQueue,
							   kAudioQueueProperty_MagicCookie,
							   magicCookie,
							   cookieSize);
		
		free (magicCookie);
		if(result < 0) goto initError;
	}
	else {
		char magicCookie[] = {	0x03, 0x19, 0x00, 0x02, 0x00, 0x04, 0x11, 0x40,
								0x15, 0x00, 0x00, 0x00, 0x00, 0x01, 0xf4, 0xf0,
								0x00, 0x00, 0x00, 0x00, 0x05, 0x02, 0x12, 0x10, 0x06, 0x01, 0x02 };
		cookieSize = 27;
		result = AudioQueueSetProperty (pState->mQueue,
										kAudioQueueProperty_MagicCookie,
										magicCookie,
										cookieSize);
		if(result < 0) goto initError;
	}

	
	result = AudioFileClose(file);
	if(result < 0)
	{
		DLog(@"MWAudioDecoder: audio file close failed with error %d", result);
	}
	
	pState->mCurrentPacket = 0;
	pState->decoder = self;
	
	// Allocate buffers
	for (int i = 0; i < kNumberBuffers; ++i)
	{
		result = AudioQueueAllocateBuffer (pState->mQueue,
										   pState->bufferByteSize,
										   &pState->mBuffers[i]);
		if(result < 0) goto initError;
	}
	
	self.inputBuffer = buf;
	
	// Initialize multiplier to convert from Host Time to seconds
	kern_return_t kerror;
	mach_timebase_info_data_t tinfo;
	kerror = mach_timebase_info(&tinfo);
	if (kerror != KERN_SUCCESS) goto initError;
	_hostTime2SecFactor = ((double)tinfo.numer / tinfo.denom)/1.0e9;
	
	return self;
	
initError:
	DLog(@"MWAudioDecoder: audio queue error %d", result);
	AudioFileClose(file);
	[self release];
	return nil;
}

-(void)prime
{
	AQPlayerState *pState = (AQPlayerState *)_state;
	pState->mIsRunning = YES;
	
	for(int i = 0; i < kNumberBuffers; i++)
	{
		HandleOutputBuffer (pState,
							pState->mQueue,
							pState->mBuffers[i]);
	}
	
	UInt32 framesPrepared = 0;
	AudioQueuePrime(pState->mQueue, 0, &framesPrepared);	
}

-(void)play
{
	AQPlayerState *pState = (AQPlayerState *)_state;
	[self setStartAudioTimeStamp];
	AudioQueueStart (pState->mQueue, NULL);
}

-(NSTimeInterval)getCurrentAudioTime
{
	NSTimeInterval time;
	Float64 sampleRate;
	AudioTimeStamp timestamp;
	@synchronized(self)
	{

		[self getCurrentAudioTimeStamp:&timestamp];
		
		if((timestamp.mFlags & kAudioTimeStampSampleTimeValid) != 0)
			sampleRate = ((AQPlayerState *)_state)->mDataFormat.mSampleRate;
	}
	AudioTimeStamp difference;
	audioTimeStampDifference(&difference, &_startTime, &timestamp);
	if((difference.mFlags & kAudioTimeStampSampleTimeValid) != 0)
	{
		time = difference.mSampleTime / sampleRate;
	}
	else
	if((difference.mFlags & kAudioTimeStampHostTimeValid) != 0)
	{
		time = (Float64)difference.mHostTime * _hostTime2SecFactor;
	}
	else
	{
		time = -100.0;
	}

	return time;
}

-(void)getCurrentAudioTimeStamp:(AudioTimeStamp *)ts
{
	if(ts == NULL) return;
	AQPlayerState *pState = (AQPlayerState *)_state;
	AudioQueueRef queue = pState->mQueue;
	OSStatus error = AudioQueueGetCurrentTime(queue, NULL, ts, NULL);
	if(error < 0)
	{
		DLog(@"MWAudioDecoder: error %d getting audio time", error);
		ts->mFlags = 0;
	}
}

-(void)setStartAudioTimeStamp
{
	_startTime.mSampleTime = 0.0;
	_startTime.mHostTime = mach_absolute_time();
	_startTime.mRateScalar = 0;
	_startTime.mWordClockTime = 0;
	memset (&_startTime.mSMPTETime, 0, sizeof (SMPTETime));
	_startTime.mFlags = kAudioTimeStampSampleTimeValid | kAudioTimeStampHostTimeValid;
}
@end

