//
//  MWVideoDecoder.m
//  videoGL
//
//  Created by Romain Criton on 28/07/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import "MWVideoDecoder.h"

#define FRAME_MODULO	30

void updateStats(MWVideoDecoderStats *stats, NSTimeInterval start, NSTimeInterval end, int frameFinished)
{
	if(stats == NULL) return;
	
	if(stats->frameStart)
	{
		if(stats->nbFrames == 0)
			stats->instantFrameRateWindowStart = start;
		else
		{
			NSTimeInterval period = start - stats->prevFrameStart;
			stats->totalPeriod += period;
			if(period > stats->maxPeriod) stats->maxPeriod = period;
			if(period < stats->minPeriod) stats->minPeriod = period;
			stats->avgPeriod = stats->totalPeriod / stats->nbFrames;
			NSTimeInterval delta = period - stats->avgPeriod;
			stats->totalSquareSumPeriod += delta * delta;
			stats->sigmaSquaredPeriod = stats->totalSquareSumPeriod / stats->nbFrames;
			stats->prevFrameStart = start;
			
			stats->instantFrameCount++;
			NSTimeInterval window = start - stats->instantFrameRateWindowStart;
			if(window > stats->instantFrameRateWindowLength)
			{
				stats->instantFrameRate = stats->instantFrameCount / window;
				stats->instantFrameCount = 0;
				stats->instantFrameRateWindowStart = start;
			}
		}

	}
	NSTimeInterval duration = end - start;
	stats->currentFrameDuration += duration;
	if(frameFinished == 0) return;

	// The frame is finished
	stats->totalDuration += stats->currentFrameDuration;
	stats->nbFrames++;
	if(stats->currentFrameDuration > stats->maxDuration) stats->maxDuration = stats->currentFrameDuration;
	if(stats->currentFrameDuration < stats->minDuration) stats->minDuration = stats->currentFrameDuration;
	stats->avgDuration = stats->totalDuration / stats->nbFrames;
	NSTimeInterval delta = stats->currentFrameDuration - stats->avgDuration;
	stats->totalSquareSumDuration += delta * delta;
	stats->sigmaSquaredDuration = stats->totalSquareSumDuration / stats->nbFrames;
	stats->currentFrameDuration = 0;
}

@implementation MWVideoDecoder

@synthesize videoSettings = _videoSettings;
@synthesize stats = _stats;

-(NSTimeInterval)pts2Timestamp:(int64_t)pts
{
	AVRational timebase = _codecCtx->time_base;
	// Timebase in Codec Context is half that in Stream (which is correct).
	// For instance: Stream: 1/15 (correct) -> Codec Context: 1/30 (double frame rate)
	NSTimeInterval timestamp = (NSTimeInterval)pts * (NSTimeInterval)timebase.num / ((NSTimeInterval)(timebase.den / 2));
	return timestamp;
}

-(id)initWithCodec:(MWMediaCodec *)codec
{
	if(codec == NULL) return nil;
	self = [super init];
	if(self == nil) return self;
	
	_codecCtx = (AVCodecContext *)codec;
	
	// Find the decoder for the video stream
    _codec = avcodec_find_decoder(_codecCtx->codec_id);
    if(_codec == NULL)
        goto initError; // Codec not found
	
    // Open codec
    if(avcodec_open2(_codecCtx, _codec, NULL) < 0)
        goto initError; // Could not open codec
	
    // Allocate video frame
    _frame = av_frame_alloc();
	
	_videoSettings.width = _codecCtx->width;
	_videoSettings.height = _codecCtx->height;
	
	// Setup scaler
	static int sws_flags =  SWS_FAST_BILINEAR;
	_scalerCtx = sws_getContext(_codecCtx->width, 
									 _codecCtx->height,
									 _codecCtx->pix_fmt,
									 _videoSettings.width, 
									 _videoSettings.height,
									 AV_PIX_FMT_RGB24,
									 sws_flags, NULL, NULL, NULL);
	
	_stats.frameStart = YES;
	_stats.instantFrameRateWindowLength = 2.0;
	_stats.minDuration = _stats.minPeriod = 1.0e9;
	_stats.totalDuration = _stats.totalDuration = 0.0;
	_stats.totalSquareSumDuration = _stats.totalSquareSumPeriod = 0.0;
	
	return self;
	
initError:
	[self release];
	return nil;
}

-(void)dealloc
{
	// Release old picture and scaler
	sws_freeContext(_scalerCtx);
	[super dealloc];
}

-(MWVideoFrame *)decodePacket:(MWMediaPacket *)packet
{
	NSTimeInterval start = [NSDate timeIntervalSinceReferenceDate];

	int frameFinished = 0;
	avcodec_decode_video2(_codecCtx, _frame, &frameFinished, packet.data);
	
	MWVideoFrame *vFrame = nil;
	if(frameFinished != 0)
	{
		void *buf = malloc(_codecCtx->width * _codecCtx->height * 4);
		int linesize[4];
		linesize[0] = 3 * _codecCtx->width;
		linesize[1] = linesize[2] = linesize[3] = 0;
		uint8_t *data[4];
		data[0] = buf;
		data[1] = data[2] = data[3] = NULL;
		sws_scale (_scalerCtx, _frame->data, _frame->linesize,
				   0, _codecCtx->height,
				   data, linesize);
		vFrame = [[MWVideoFrame alloc] initWithData:buf width:_codecCtx->width height:_codecCtx->height];
		vFrame.timestamp = [self pts2Timestamp:((AVPacket *)packet.data)->pts]; 
	}
	
	NSTimeInterval end = [NSDate timeIntervalSinceReferenceDate];
	updateStats(&_stats, start, end, frameFinished);
	return [vFrame autorelease];
}

-(NSTimeInterval)getTimeStampForPacket:(MWMediaPacket *)packet
{
	if((packet == nil) || (packet.data == NULL)) return -1.0;
	
	return [self pts2Timestamp:((AVPacket *)packet.data)->pts];
}

@end
